#ifndef _FPGA_I2C_H
#define _FPGA_I2C_H

#include "Prefix.h"
#include "OS_Semaphore.h"
#include "OS_Thread.h"
#include "Bus_I2C.h"
#include "FPGA_Register.h"
#include "EARTH_PT3.h"
#include <time.h>

namespace EARTH {
namespace FPGA {
	class I2C {
	public:
		I2C(FPGA::Register *register_) : mRegister(register_)
		{
			_ASSERT(mRegister);

			mDataPtr = NULL;
			mSemaphore = NULL;

			_ASSERT(mSemaphore == NULL);
			mSemaphore = new OS::Semaphore(1);
			_ASSERT(mSemaphore);
		}

		~I2C()
		{
			delete_(mSemaphore);
		}

		void Init()
		{
			volatile void	*ptr;
			mRegister->GetBaseAddress(1, &ptr);
			mDataPtr = static_cast<volatile uint8 *>(ptr) + 2*1024;

//			memset(const_cast<void *>(ptr), 0, 4096);

//			Test();
		}

		void Copy(Bus::I2C *bus)
		{
			_ASSERT(bus);
			_ASSERT(bus->IsEnd());

			const std::vector<uint8> &instructions = bus->mInstructions;
			
			const uint8		*src = &instructions[0];
			uint8			*dst = const_cast<uint8 *>(mDataPtr + bus->InstructionAddress()/2);
			::memcpy(dst, src, instructions.size());
		}

		status Run(Bus::I2C *bus, uint32 *ack = NULL, bool copy = true)
		{
			_ASSERT(bus);
			_ASSERT(bus->IsEnd());

			_ASSERT(mSemaphore);
			mSemaphore->Acquire();

			if (copy) {
				Copy(bus);
			}

			std::vector<uint8>	readDataVector;
			status status = RunCode(bus->InstructionAddress(), ack);

			uint32	readSize = bus->ReadAddress();

			for (uint32 i=0; i<readSize; i++) {
				bus->PushReadData(mDataPtr[i]);
			}

			_ASSERT(mSemaphore);
			mSemaphore->Release();

			return status;
		}

		void GetClean(bool *b)
		{
			uint32	data;

			_ASSERT(mRegister);
			mRegister->Read32(0x14, &data);

			_ASSERT(b);
			*b = BIT_SHIFT_MASK(data, 3, 1) ? true : false;
		}

		void Reset()
		{
			_ASSERT(mRegister);
			mRegister->Write32(0x10, 1<<17);
		}

	protected:
		friend class I2C_;

		FPGA::Register	*const mRegister;
		volatile uint8	*mDataPtr;
		OS::Semaphore	*mSemaphore;

		enum {
			SIZE_INSTRUCTION = 2*1024,	// CXgNV̈: 2koCg
			SIZE_DATA        = 2*1024,	// f[^̈            : 2koCg
		};

		status RunCode(uint32 startAddress, uint32 *ack_)
		{
			// Run O Busy NA܂ő҂
			uint32	data;
			Wait(&data);

			_ASSERT(startAddress < (1<<13));
			mRegister->Write32(0x10, 1<<16 | startAddress);

			Wait(&data);

			uint32	ack = BIT_SHIFT_MASK(data, 1, 2);
			if (ack_) {
				*ack_ = ack;
			}

			bool	error = BIT_SHIFT_MASK(data, 1, 2) ? true : false;

			return error ? PT::STATUS_I2C_ERROR : PT::STATUS_OK;
		}

		void Wait(uint32 *data_ = NULL)
		{
			uint32	data;
			while (true) {
				mRegister->Read32(0x14, &data);
				if (!BIT_SHIFT_MASK(data, 0, 1)) break;
				OS::Thread::Sleep(0);
			}

			if (data_) {
				*data_ = data;
			}
		}

		uint32 DataN(uint32 index, uint32 n) const
		{
			_ASSERT((1 <= n) && (n <= 4));

			uint32 data = 0;

			for (uint32 i=0; i<n; i++) {
				data <<= 8;
				data  |= mDataPtr[index + i];
			}

			return data;
		}

		uint8 Data1(uint32 index) const
		{
			return static_cast<uint8>(DataN(index, 1));
		}

		uint16 Data2(uint32 index) const
		{
			return static_cast<uint16>(DataN(index, 2));
		}

		uint32 Data3(uint32 index) const
		{
			return DataN(index, 3);
		}

		uint32 Data4(uint32 index) const
		{
			return DataN(index, 4);
		}
	};
}
}

#endif
