#ifndef _MxL301RF_H
#define _MxL301RF_H

#include "Prefix.h"
#include "TC90522_T.h"
#include "Tuner_T.h"
#include "Bus_I2C.h"
#include "OS_Time.h"

#define I2C_NULL NULL

namespace EARTH {
	class MxL301RF : public Tuner_T {
	public:
		static uint8 Address(uint32 index)
		{
			const uint8		ADDRESS[2] = { 0x62, 0x61 };
			_ASSERT(index < sizeof(ADDRESS)/sizeof(*ADDRESS));
			return ADDRESS[index];
		}

		MxL301RF(TC90522_T *tc90522) : Tuner_T(tc90522)
		{
			_ASSERT(mTC90522);

			mSleep = true;

			mChannel = 0;
			mOffset  = 0;

			mWriteFunction = NULL;
			mReadFunction  = NULL;
			mSleepFunction = NULL;
		}

		~MxL301RF()
		{
		}

		status SetSleep(bool sleep)
		{
			status	status;

			if (sleep) {
				status = mTC90522->SetAGC(I2C_NULL, EARTH::TC90522::AGC_MANUAL); CHECK_STATUS;
				SetSleepMode(I2C_NULL, sleep);
				mTC90522->Write_slptim(I2C_NULL, sleep);
			} else {
				mTC90522->Write_slptim(I2C_NULL, sleep);
				SetSleepMode(I2C_NULL, sleep);
			}

			mSleep = sleep;

			return PT::STATUS_OK;
		}

		void GetSleep(bool *sleep) const
		{
			_ASSERT(sleep);
			*sleep = mSleep;
		}

		status SetFrequency(uint32 channel, sint32 offset)
		{
			status	status;

			status = mTC90522->SetAGC(I2C_NULL, EARTH::TC90522::AGC_MANUAL); CHECK_STATUS;

			bool	catv;
			uint32	number, frequency;
			GetChannelFrequency(channel, &catv, &number, &frequency);

			double	realFrequency = (7*frequency+1+offset)*1000000.0/7.0;

			SetFrequency(I2C_NULL, static_cast<uint32>(realFrequency));

			uint32	startTime;
			OS::Time::Get(&startTime);

			bool	locked1 = false,
					locked2 = false;
			while (true) {
				uint32	nowTime;
				OS::Time::Get(&nowTime);

				GetLocked1(I2C_NULL, &locked1);
				GetLocked2(I2C_NULL, &locked2);
				if (locked1 && locked2) break;

				uint32	deltaTime = nowTime - startTime;
				if (100 <= deltaTime) break;

				OS::Thread::Sleep(1);
			}
			if (!(locked1 && locked2)) return PT::STATUS_PLL_LOCK_TIMEOUT_ERROR;

			status = mTC90522->SetAGC(I2C_NULL, EARTH::TC90522::AGC_AUTO); CHECK_STATUS;

			return status;
		}

		void GetFrequency(uint32 *channel, sint32 *offset) const
		{
			*channel = mChannel;
			*offset  = mOffset;
		}

		status GetLocked1(Bus::I2C *bus, bool *locked)
		{
			MxL_RFSynth_Lock_Status(bus, locked);

			return PT::STATUS_OK;
		}

		status GetLocked2(Bus::I2C *bus, bool *locked)
		{
			MxL_REFSynth_Lock_Status(bus, locked);

			return PT::STATUS_OK;
		}

		status GetLevel(Bus::I2C *bus, float *level)
		{
			MxL_Check_RF_Input_Power(bus, level);

			return PT::STATUS_OK;
		}

	protected:
		friend class MxL301RF_;

		uint32		mFreq;

		bool		mSleep;

		uint32		mChannel;
		sint32		mOffset;

		typedef void (*WriteFunction)(const uint8 *, uint32, void *);
		typedef void (*ReadFunction )(uint8, uint8 *, void *);
		typedef void (*SleepFunction)(uint32, void *);

		WriteFunction	mWriteFunction;
		ReadFunction	mReadFunction;
		SleepFunction	mSleepFunction;
		void			*mFunctionArg;

		void SetCallbackFunctions(WriteFunction write, ReadFunction read, SleepFunction sleep, void *arg)
		{
			mWriteFunction = write;
			mReadFunction  = read;
			mSleepFunction = sleep;
			mFunctionArg   = arg;
		}

		void MxL_Set_Register(Bus::I2C *bus, uint8 RegAddr, uint8 RegData)
		{
			uint8	data[2];
			data[0] = RegAddr;
			data[1] = RegData;
	
			Write(bus, data, sizeof(data));
		}

		void MxL_Get_Register(Bus::I2C *bus, uint8 RegAddr, uint8 *RegData)
		{
			Read(bus, RegAddr, RegData);
		}

		status SetSleepMode(Bus::I2C *bus, bool sleep)
		{
			status	status = 0;

			if (sleep) {
				MxL_Stand_By(bus);
			} else {
				MxL_Wake_Up(bus);
			}

			return status;
		}

		void SetFrequency(Bus::I2C *bus, uint32 frequency)
		{
			MxL_Tuner_RFTune(bus, frequency);
		}

		void MxL_Stand_By(Bus::I2C *bus)
		{
			uint8	data[4];	/* a array pointer that store the addr and data pairs for I2C write	*/
	
			data[0] = 0x01;
			data[1] = 0x00;
			data[2] = 0x13;
			data[3] = 0x00;

			Write(bus, data, sizeof(data));
		}

		void MxL_Wake_Up(Bus::I2C *bus)
		{
			uint8	data[2];	/* a array pointer that store the addr and data pairs for I2C write	*/

			data[0] = 0x01;
			data[1] = 0x01;

			Write(bus, data, sizeof(data));

			MxL_Tuner_RFTune(bus, mFreq);
		}

		void MxL_idac_setting(Bus::I2C *bus)
		{
			uint8 Array[] = {
				0x0D, 0x00,
				0x0C, 0x67,
				0x6F, 0x89,
				0x70, 0x0C,
				0x6F, 0x8A,
				0x70, 0x0E,
				0x6F, 0x8B,
				0x70, 0x10+12,
			};

			Write(bus, Array, sizeof(Array));
		}

		void MxL301RF_RFTune(uint8 *pArray, uint32 *Array_Size, uint32 RF_Freq)
		{
			uint8	data[]=
			{
				0x13, 0x00,		// abort tune
				0x3B, 0xC0,
				0x3B, 0x80,
				0x10, 0x95,		// BW
				0x1A, 0x05,
				0x61, 0x00,
				0x62, 0xA0,
				0x11, 0x40,		// 2 bytes to store RF frequency
				0x12, 0x0E,		// 2 bytes to store RF frequency
				0x13, 0x01		// start tune
			};

			uint32	dig_rf_freq=0;
			uint32	temp = 0 ;
			uint32	frac_divider = 1000000;
			uint32	kHz = 1000;

			enum {
				MHz = 1000000
			};

			// Convert RF frequency into 16 bits => 10 bit integer (MHz) + 6 bit fraction
			dig_rf_freq = RF_Freq / MHz;	// Whole number portion of RF freq (in MHz)
			temp = RF_Freq % MHz;			// Decimal portion of RF freq (in MHz)
			for (uint32 i=0; i<6; i++) {
				dig_rf_freq <<= 1;
				frac_divider /=2;
				if(temp > frac_divider) {	// Carryover from decimal
					temp -= frac_divider;
					dig_rf_freq++;
				}
			}

			// add to have shift center point by 7.8124 kHz
			if (temp > 7812) {
				dig_rf_freq ++;
			}

			data[2*(7)+1] = static_cast<uint8>(dig_rf_freq     );
			data[2*(8)+1] = static_cast<uint8>(dig_rf_freq >> 8);

			struct SHFType {
				uint32	Freq;		// Channel center frequency
				uint32	Freq_TH;	// Offset frequency threshold
				uint8	SHF_Val;	// Spur shift value
				uint8	SHF_Dir;	// Spur shift direction
			};

			const SHFType SHF_DVBT_TAB[] = {
				// { Freq(kHz),Offset(kHz), Val,	Dir,	type},
				{  64500, 500, 0x92, 0x07 },
				{ 191500, 300, 0xE2, 0x07 },
				{ 205500, 500, 0x2C, 0x04 },
				{ 212500, 500, 0x1E, 0x04 },
				{ 226500, 500, 0xD4, 0x07 },
				{  99143, 500, 0x9C, 0x07 },
				{ 173143, 500, 0xD4, 0x07 },
				{ 191143, 300, 0xD4, 0x07 },
				{ 207143, 500, 0xCE, 0x07 },
				{ 225143, 500, 0xCE, 0x07 },
				{ 243143, 500, 0xD4, 0x07 },
				{ 261143, 500, 0xD4, 0x07 },
				{ 291143, 500, 0xD4, 0x07 },
				{ 339143, 500, 0x2C, 0x04 },
				{ 117143, 500, 0x7A, 0x07 },
				{ 135143, 300, 0x7A, 0x07 },
				{ 153143, 500, 0x01, 0x07 }
			};

			// Spur shift function for analog mode V9.2.7.0
			for (uint32 i=0; i<sizeof(SHF_DVBT_TAB)/sizeof(*SHF_DVBT_TAB); i++) {
				if (   (RF_Freq >= (SHF_DVBT_TAB[i].Freq - SHF_DVBT_TAB[i].Freq_TH)*kHz)
					&& (RF_Freq <= (SHF_DVBT_TAB[i].Freq + SHF_DVBT_TAB[i].Freq_TH)*kHz)) {
					data[2*(5)+1] = SHF_DVBT_TAB[i].SHF_Val;
					data[2*(6)+1] = 0xa0 | SHF_DVBT_TAB[i].SHF_Dir;
					break;
				}
			}

			::memcpy(pArray, data, sizeof(data));
			
			*Array_Size = sizeof(data);
		}

		void MxL_Tuner_RFTune(Bus::I2C *bus, uint32 RF_Freq_Hz)
		{
			uint8	data[100];
			uint32	size = 0;

			mFreq = RF_Freq_Hz;

			// perform Channel Change calculation 
			MxL301RF_RFTune(data,&size, RF_Freq_Hz);

			_ASSERT(size == 20);

			Write(bus, data, 14);

			Sleep(3);

			Write(bus, data + 14, 6);

			Sleep(1);
			Sleep(30);

			MxL_Set_Register(bus, 0x1A, 0x0D);

			MxL_idac_setting(bus);
		}

		void MxL_RFSynth_Lock_Status(Bus::I2C *bus, bool *locked)
		{
			_ASSERT(locked);
			*locked = false; 

			uint8 Data;
			MxL_Get_Register(bus, 0x16, &Data);

			Data &= 0x0C;
			if (Data == 0x0C) {
				*locked = true;		// RF Synthesizer is Lock
			}
		}

		void MxL_REFSynth_Lock_Status(Bus::I2C *bus, bool *locked)
		{
			_ASSERT(locked);
			*locked = false; 

			uint8	Data;
			MxL_Get_Register(bus, 0x16, &Data);

			Data &= 0x03;
			if (Data == 0x03) {
				*locked = true;   // REF Synthesizer is Lock
			}
		}

		void MxL_Check_RF_Input_Power(Bus::I2C *bus, float *RF_Input_Level)
		{
			uint8	RFin1, RFin2, RFOff1, RFOff2;
			float	RFin, RFoff;
			float	cal_factor;

			MxL_Set_Register(bus, 0x14, 0x01);

			Sleep(1);
			MxL_Get_Register(bus, 0x18, &RFin1);
			MxL_Get_Register(bus, 0x19, &RFin2);

			MxL_Get_Register(bus, 0xD6, &RFOff1);
			MxL_Get_Register(bus, 0xD7, &RFOff2);

			RFin  = (float)(((RFin2 & 0x07) << 5) + ((RFin1 & 0xF8) >> 3) + ((RFin1 & 0x07) * 0.125));
			RFoff = (float)(((RFOff2 & 0x0F) << 2) + ((RFOff1 & 0xC0) >> 6) + (((RFOff1 & 0x38)>>3) * 0.125));

			cal_factor = 113.;

			*RF_Input_Level = RFin - RFoff - cal_factor;
		}

		void Write(Bus::I2C *bus, const uint8 *data, uint32 size)
		{
			if (mWriteFunction) {
				mWriteFunction(data, size, mFunctionArg);
			} else {
				_ASSERT(mTC90522);
				mTC90522->WriteTunerWithoutAddress(bus, data, size);
			}
		}

		void Read(Bus::I2C *bus, uint8 address, uint8 *data)
		{
			if (mReadFunction) {
				mReadFunction(address, data, mFunctionArg);
			} else {
				uint8	write[2];
				write[0] = 0xfb;
				write[1] = address;

				_ASSERT(mTC90522);
				mTC90522->WriteTunerWithoutAddress(bus, write, sizeof(write));
				mTC90522->ReadTunerWithoutAddress(bus, data, 1);
			}
		}

		void Sleep(uint32 ms)
		{
			if (mSleepFunction) {
				mSleepFunction(ms, mFunctionArg);
			} else {
				OS::Thread::Sleep(ms);
			}
		}
	};
}

#endif
