#ifndef _TC90522_T_H
#define _TC90522_T_H

#include "Prefix.h"
#include "FPGA_I2C.h"
#include "EARTH_PT3.h"
#include "TC90522.h"
#include <math.h>

namespace EARTH {
	class TC90522_T : public TC90522 {
	public:
		TC90522_T(FPGA::I2C *i2c, uint8 myAddress, uint8 tunerAddress)
			: TC90522(i2c, myAddress, tunerAddress)
		{
			mLnbPower = PT::Device::LNB_POWER_OFF;
			mTunerPower = true;
			mAmpPower   = false;
		}

		~TC90522_T()
		{
		}

		status Init(Bus::I2C *bus)
		{
			status	status;

			status = Write_imsrst(bus); CHECK_STATUS;
			status = Write(bus, 0x1c, 0x10); CHECK_STATUS;

			return status;
		}

		status SetTsPinsMode(Bus::I2C *bus, const PT::Device::TsPinsMode *mode)
		{
			uint32	clock_data = mode->clock_data,
					byte       = mode->byte,
					valid      = mode->valid;

			if (clock_data) clock_data++;
			if (byte      ) byte++;
			if (valid     ) valid++;

			uint8	data = {
				static_cast<uint8>(0x01 | clock_data << 6 | byte << 4 | valid << 2)
			};

			status	status;

			status = Write(bus, 0x1d, data); CHECK_STATUS;

			return status;
		}

		status SetAGC(Bus::I2C *bus, AGC agc)
		{
			status	status;

			status = Write(bus, 0x25, (agc == AGC_AUTO) ? 64 : 0); CHECK_STATUS;

			uint8	data;
			data  = 0x4c;
			data |= (agc == AGC_AUTO) ? 0x00 : 0x01;
			status = Write(bus, 0x23, data); CHECK_STATUS;

			status = Write_imsrst(bus); CHECK_STATUS;

			return status;
		}

		PT::Device::LnbPower	mLnbPower;
		bool					mTunerPower;
		bool					mAmpPower;

		status GetLnbPower(PT::Device::LnbPower *lnbPower)
		{
			*lnbPower = mLnbPower;

			return PT::STATUS_OK;
		}

		status SetTunerPower(Bus::I2C *bus, bool b)
		{
			mTunerPower = b;

			status	status;

			status = SetPowers(bus); CHECK_STATUS;

			return status;
		}

		void SetTunerPower(bool b)
		{
			mTunerPower = b;
		}

		void SetAmpPower(bool b)
		{
			mAmpPower = b;
		}

		status SetPowers(Bus::I2C *bus)
		{
			uint8	tunerPower = mTunerPower ? BIT_2(1,1) : BIT_2(1,0),
					ampPower   = mAmpPower   ? BIT_2(1,1) : BIT_2(1,0);

			uint8 data = tunerPower << 6 | BIT_2(0,1) << 4 | ampPower << 2 | BIT_2(0,1) << 0;

			status	status;

			status = Write(bus, 0x1e, data); CHECK_STATUS;

			return status;
		}

		// ISDB-T
		status Write_slptim(Bus::I2C *bus, bool sleep)
		{
			status	status;

			uint8	data = BIT_1(1)<<7 | ((sleep ? 1 : 0) << 4);
			status = Write(bus, 0x03, data); CHECK_STATUS;

			return status;
		}

		status Write_mlocksel(Bus::I2C *bus, bool search)
		{
			status	status;

			uint8	data = (search ? 1 : 0) << 6;
			status = Write(bus, 0x47, data); CHECK_STATUS;

			return status;
		}

		status Write_laysel(Bus::I2C *bus, uint8 layer)
		{
			layer ^= BIT_3(1,1,1);
			uint8	data = 0;
			if (BIT_SHIFT_MASK(layer, 0, 1)) data |= BIT_1(1) << 2;
			if (BIT_SHIFT_MASK(layer, 1, 1)) data |= BIT_1(1) << 1;
			if (BIT_SHIFT_MASK(layer, 2, 1)) data |= BIT_1(1) << 0;

			status	status;
			
			status = Write(bus, 0x71, data); CHECK_STATUS;

			return status;
		}

		status Write_perst(Bus::I2C *bus)
		{
			status	status;

			status = Write(bus, 0x76, 0x02 | BIT_1(1) << 0); CHECK_STATUS;

			return status;
		}

		status Read_eqqth(Bus::I2C *bus, uint32 *eqqth)
		{
			status	status;

			uint8	data;
			status = Read(bus, 0x4f, &data); CHECK_STATUS;

			_ASSERT(eqqth);
			*eqqth = BIT_SHIFT_MASK(data, 5, 3);

			return status;
		}

		status Read_retryov_tmunvld_fulock(Bus::I2C *bus, bool *retryov, bool *tmunvld, bool *fulock)
		{
			status	status;

			uint8	data;
			status = Read(bus, 0x80, &data); CHECK_STATUS;

			_ASSERT(retryov); *retryov = BIT_SHIFT_MASK(data, 7, 1) ? true : false;
			_ASSERT(tmunvld); *tmunvld = BIT_SHIFT_MASK(data, 5, 1) ? true : false;
			_ASSERT(fulock ); *fulock  = BIT_SHIFT_MASK(data, 3, 1) ? true : false;

			return status;
		}

		status Read_ifagc_dt(Bus::I2C *bus, uint8 *ifagc_dt)
		{
			status	status;

			_ASSERT(ifagc_dt);
			status = Read(bus, 0x82, ifagc_dt); CHECK_STATUS;
			
			return status;
		}

		status Read_carafc_dt(Bus::I2C *bus, sint32 *hz)
		{
			status	status;

			uint8	data[2];
			status = Read(bus, 0x84, data, sizeof(data)); CHECK_STATUS;

			sint16	carafc_dt = Byte2(data);

			double	a = carafc_dt * 7.75 * mMasterClockFrequencyMHz / 65.015873;

			_ASSERT(hz);
			*hz = static_cast<sint32>(a + 0.5);

			return status;
		}

		status Read_clkafc_dt(Bus::I2C *bus, sint32 *ppm100)
		{
			status	status;

			uint8	data[2];
			status = Read(bus, 0x86, data, sizeof(data));

			sint16	short_data = Byte2(data);

			sint32	clkafc_dt = static_cast<sint32>(short_data);

			const	sint32 HKFRQ[2] = { 11274, 17039 };

			sint32	hkfrq = HKFRQ[0/*mXT*/];
			sint32	hkncog = 1;

			double	a = hkfrq + ::pow(2, static_cast<double>(16));
			double	b = hkfrq + clkafc_dt / ::pow(2, static_cast<double>(10+hkncog)) + ::pow(2, static_cast<double>(16));

			_ASSERT(ppm100);
			*ppm100 = static_cast<sint32>((1 - a / b) * ::pow(10, static_cast<double>(6+2)) + 0.5);

			return status;
		}

		status Read_cndat(Bus::I2C *bus, uint32 *cn)
		{
			status	status;

			uint8	data[3];
			status = Read(bus, 0x8b, data, sizeof(data)); CHECK_STATUS;

			uint32	pi = Byte3(data);

			double p;
			if (pi) {
				p = pi;
				p = 10*log10(5505024 / p);

				double p2 = ::pow(p, 2);
				double p3 = ::pow(p, 3);
				double p4 = ::pow(p, 4);

				p = 0.000024*p4 - 0.0016*p3 + 0.0398*p2 + 0.5491*p + 3.0965;
			} else {
				p = 0;
			}

			p *= 100;

			_ASSERT(cn);
			*cn = static_cast<uint32>(p + 0.5);

			return status;
		}

		status Read_rlock(Bus::I2C *bus, bool *a, bool *b, bool *c)
		{
			status	status;

			uint8	data;
			status = Read(bus, 0x96, &data); CHECK_STATUS;

			_ASSERT(a); *a = BIT_SHIFT_MASK(data, 7, 1);
			_ASSERT(b); *b = BIT_SHIFT_MASK(data, 6, 1);
			_ASSERT(c); *c = BIT_SHIFT_MASK(data, 5, 1);

			return status;
		}

		status Read_perr_pecy(Bus::I2C *bus, PT::Device::LayerIndex layerIndex, uint32 *numerator, uint32 *denominator)
		{
			status	status;

			uint8	perr[3];
			uint8	pecy[2];

			status = Read(bus, static_cast<uint8>(0x9d+sizeof(perr)*layerIndex), perr, sizeof(perr)); CHECK_STATUS;
			status = Read(bus, static_cast<uint8>(0xa6+sizeof(pecy)*layerIndex), pecy, sizeof(pecy)); CHECK_STATUS;

			_ASSERT(numerator);
			*numerator = Byte3(perr);

			_ASSERT(denominator);
			*denominator = 8 * 204 * Byte2(pecy);

			return status;
		}

		status Read_ffsize_gdleng_seqen(Bus::I2C *bus, uint32 *ffsize, uint32 *gdleng, uint32 *seqen)
		{
			status	status;

			uint8	data;
			status = Read(bus, 0xb0, &data); CHECK_STATUS;

			_ASSERT(ffsize); *ffsize = BIT_SHIFT_MASK(data, 6, 2);
			_ASSERT(gdleng); *gdleng = BIT_SHIFT_MASK(data, 4, 2);
			_ASSERT(seqen  ); *seqen = BIT_SHIFT_MASK(data, 0, 4);

			return status;
		}

		status Read_TMCC(Bus::I2C *bus, PT::Device::TmccT *tmcc)
		{
			status	status;

			uint8	data[8];
			status = Read(bus, 0xb2+0, data+0, 4); CHECK_STATUS;
			status = Read(bus, 0xb2+4, data+4, 4); CHECK_STATUS;

			// 0:b2 1:b3 2:b4 3:b5 4:b6 5:b7 6:b8 7:b9
			tmcc->System    = BIT_SHIFT_MASK(data[0], 6, 2);
			tmcc->Indicator = BIT_SHIFT_MASK(data[0], 2, 4);
			tmcc->Emergency = BIT_SHIFT_MASK(data[0], 1, 1);
			tmcc->Partial   = BIT_SHIFT_MASK(data[0], 0, 1);

			tmcc->Mode[0] = BIT_SHIFT_MASK(data[1], 5, 3);
			tmcc->Mode[1] = BIT_SHIFT_MASK(data[2], 0, 3);
			tmcc->Mode[2] = BIT_SHIFT_MASK(data[4], 3, 3);

			tmcc->Rate[0] = BIT_SHIFT_MASK(data[1], 2, 3);
			tmcc->Rate[1] = BIT_SHIFT_MASK(data[3], 5, 3);
			tmcc->Rate[2] = BIT_SHIFT_MASK(data[4], 0, 3);

			uint32	interleave0h = BIT_SHIFT_MASK(data[1], 0, 2);
			uint32	interleave0l = BIT_SHIFT_MASK(data[2], 7, 1);

			tmcc->Interleave[0] = interleave0h << 1 | interleave0l << 0;
			tmcc->Interleave[1] = BIT_SHIFT_MASK(data[3], 2, 3);
			tmcc->Interleave[2] = BIT_SHIFT_MASK(data[5], 5, 3);

			uint32	segment1h = BIT_SHIFT_MASK(data[3], 0, 2);
			uint32	segment1l = BIT_SHIFT_MASK(data[4], 6, 2);

			tmcc->Segment[0] = BIT_SHIFT_MASK(data[2], 3, 4);
			tmcc->Segment[1] = segment1h << 2 | segment1l << 0;
			tmcc->Segment[2] = BIT_SHIFT_MASK(data[5], 1, 4);

			return status;
		}

		status Read_BeforeViterbi(Bus::I2C *bus, uint32 *count)
		{
			status	status;

			uint8	data[3];
			status = Read(bus, 0x97, data, sizeof(data)); CHECK_STATUS;

			uint32	data3 = Byte3(data);

			_ASSERT(count);
			*count = data3;

			return status;
		}
	protected:
		status Write_imsrst(Bus::I2C *bus)
		{
			status	status;

			status = Write(bus, 0x01, BIT_1(1) << 6); CHECK_STATUS;

			return status;
		}


	};
}

#endif
