#ifndef _PCI_Device_BIT_H
#define _PCI_Device_BIT_H

#include "PCI_Device.h"
#include "PCI.h"
#include "EARTH_PT3.h"

#include <windows.h>
#include "windrvr.h"

#if defined(KERNEL_32BIT) || defined(KERNEL_64BIT)

namespace EARTH {
namespace PCI {
namespace KERNEL_BIT {

	using namespace EARTH;

	class _Device : public Device {
	public:	
		_Device(HANDLE handle, const Target *target) : mHandle(handle)
		{
			_ASSERT(mHandle != INVALID_HANDLE_VALUE);

			mBusItem = -1;
			mMemoryItem[0] = -1;
			mMemoryItem[1] = -1;
			BZERO(mReg);
			mNeedsUnregister = false;
			
			for (uint32 i=0; i<sizeof(mData)/sizeof(*mData); i++) {
				mData[i] = 0;
			}

			_ASSERT(target);
			mSlot.dwBus      = target->Bus;
			mSlot.dwSlot     = target->Slot;
			mSlot.dwFunction = target->Function;
		}

		virtual ~_Device()
		{
			if (mNeedsUnregister) {
				_ASSERT(mHandle != INVALID_HANDLE_VALUE);
				DWORD	status = WD_CardUnregister(mHandle, &mReg);
				_ASSERT(status == WD_STATUS_SUCCESS);
				mNeedsUnregister = false;
			}
		}

		status Init()
		{
			WD_PCI_CARD_INFO	card;
			BZERO(card);
			card.pciSlot = mSlot;

			status	status;

			status = WD_PciGetCardInfo(mHandle, &card);
			if (status != WD_STATUS_SUCCESS) return PT::STATUS_WD_PciGetCardInfo_ERROR;

			for (uint32 i=0; i<card.Card.dwItems; i++) {
				const WD_ITEMS	*item = &card.Card.Item[i];
				_ASSERT(item);

				switch (item->item) {
					case ITEM_BUS:
					{
						if (0 <= mBusItem) return PT::STATUS_WD_PciGetCardInfo_Bus_ERROR;
						mBusItem = i;
						break;
					}
					case ITEM_MEMORY:
					{
						if (mMemoryItem[0] < 0) {
							mMemoryItem[0] = i;
						} else if (mMemoryItem[1] < 0) {
							mMemoryItem[1] = i;
						} else {
							return PT::STATUS_WD_PciGetCardInfo_Memory_ERROR;
						}
						break;
					}
				}
			}

			if (mBusItem < 0) return PT::STATUS_WD_PciGetCardInfo_Bus_ERROR;
			if (mMemoryItem[0] < 0) return PT::STATUS_WD_PciGetCardInfo_Memory_ERROR;

			BZERO(mReg);
			mReg.Card            = card.Card;
			mReg.fCheckLockOnly  = FALSE;
			mReg.dwOptions       = 0;
			mReg.cName[0]        = 0;
			mReg.cDescription[0] = 0;

			status = WD_CardRegister(mHandle, &mReg);
			if (status != WD_STATUS_SUCCESS) return PT::STATUS_WD_CardRegister_ERROR;
			mNeedsUnregister = true;

			return PT::STATUS_OK;
		}

		virtual void Delete()
		{
			delete this;
		}

		virtual status WriteConfig(uint32 address, const void *data, uint32 size)
		{
			WD_PCI_CONFIG_DUMP	config;
			BZERO(config);
			config.pciSlot  = mSlot;
			config.pBuffer  = const_cast<void *>(data);
			config.dwOffset = address;
			config.dwBytes  = size;
			config.fIsRead  = FALSE;

			DWORD	status = WD_PciConfigDump(mHandle, &config);
			if ((status != WD_STATUS_SUCCESS) || (config.dwResult != PCI_ACCESS_OK)) return PT::STATUS_WD_PciConfigDump_ERROR;

			return PT::STATUS_OK;
		}

		virtual status ReadConfig(uint32 address, void *data, uint32 size)
		{
			WD_PCI_CONFIG_DUMP	config;
			BZERO(config);
			config.pciSlot  = mSlot;
			config.pBuffer  = data;
			config.dwOffset = address;
			config.dwBytes  = size;
			config.fIsRead  = TRUE;

			DWORD	status = WD_PciConfigDump(mHandle, &config);
			if ((status != WD_STATUS_SUCCESS) || (config.dwResult != PCI_ACCESS_OK)) return PT::STATUS_WD_PciConfigDump_ERROR;

			return PT::STATUS_OK;
		}

		virtual void GetBaseAddress(uint32 index, volatile void **ptr_)
		{
			volatile void	*ptr = reinterpret_cast<volatile void *>(mReg.Card.Item[mMemoryItem[index]].I.Mem.dwUserDirectAddr);
			_ASSERT(ptr_);
			*ptr_ = ptr;
		}

		virtual status SetReset(uint32 data)
		{
			uint32	TABLE[6][2] = {
				{ 0x08           , 0       },
				{ 0x10           , 1 << 17 },
				{ 0x48 + 0x18*(0), 1 <<  1 },
				{ 0x48 + 0x18*(1), 1 <<  1 },
				{ 0x48 + 0x18*(2), 1 <<  1 },
				{ 0x48 + 0x18*(3), 1 <<  1 }
			};

			TABLE[0][1] = data;

			WD_TRANSFER		transfer[6];

			for (uint32 i=0; i<sizeof(transfer)/sizeof(*transfer); i++) {
				WD_TRANSFER *ptr = &transfer[i];
				ptr->dwPort     = mReg.Card.Item[mMemoryItem[0]].I.Mem.dwTransAddr + TABLE[i][0];
				ptr->cmdTrans   = WM_DWORD;
				ptr->dwBytes    = 4;
				ptr->fAutoinc   = FALSE;
				ptr->dwOptions  = 0;
				ptr->Data.Dword = TABLE[i][1];
			}

			WD_CARD_CLEANUP clean;
			BZERO(clean);

			clean.hCard     = mReg.hCard;
			clean.Cmd       = transfer;
			clean.dwCmds    = sizeof(transfer)/sizeof(*transfer);
			clean.dwOptions = 0;

			uint32 status = WD_CardCleanupSetup(mHandle, &clean);
			if (status != WD_STATUS_SUCCESS) return PT::STATUS_WD_CardCleanupSetup_ERROR;

			return PT::STATUS_OK;
		}

		virtual uint32 ID() const
		{
			return mReg.hCard;
		}

	protected:
		const HANDLE		mHandle;
		WD_PCI_SLOT			mSlot;
		sint32				mBusItem,
							mMemoryItem[2];
		WD_CARD_REGISTER	mReg;
		bool				mNeedsUnregister;
		uint32				mData[2];
	};

	status NewDevice(HANDLE handle, const Target *target, Device **device_)
	{
		_ASSERT(target);
		Device	*device = new _Device(handle, target);
		status	status = device->Init();
		if (status) {
			device->Delete();
			device = NULL;
		}

		_ASSERT(device_);
		*device_ = device;

		return status;
	}
}
}
}

#endif

#endif
